(*----------------------------------------------------------------------------*
 *  Direct3D sample from DirectX 9.0 SDK Summer 2003                          *
 *  Delphi adaptation by Alexey Barkovoy (e-mail: clootie@ixbt.com)           *
 *                                                                            *
 *  Delphi versions 5-7 are supported                                         *
 *                                                                            *
 *  Modified: 30-Dec-2003                                                     *
 *                                                                            *
 *  Latest version can be downloaded from:                                    *
 *     http://clootie.narod.ru/delphi                                         *
 *----------------------------------------------------------------------------*)
//-----------------------------------------------------------------------------
// File: PointSprites.cpp
//
// Desc: Sample showing how to use point sprites to do particle effects
//
// Copyright (c) Microsoft Corporation. All rights reserved.
//-----------------------------------------------------------------------------

unit PointSpritesUnit;

interface

uses
  Windows, Messages, SysUtils,
  Direct3D9, D3DX9, D3DApp, D3DEnumeration, D3DSettings,
  D3DFile, D3DFont, D3DUtil, DXUtil;

// Helper function to stuff a FLOAT into a DWORD argument
function FtoDW(f: Single): DWORD;

//-----------------------------------------------------------------------------
// Custom vertex types
//-----------------------------------------------------------------------------
type
  PColorVertex = ^TColorVertex;
  TColorVertex = packed record
    v: TD3DXVector3;
    color: TD3DColor;
    tu: Single;
    tv: Single;
  end;
  PColorVertexArray = ^TColorVertexArray;
  TColorVertexArray = array [0..0] of TColorVertex;

const
  TColorVertex_FVF = (D3DFVF_XYZ or D3DFVF_DIFFUSE or D3DFVF_TEX1);

type
  PPointVertex = ^TPointVertex;
  TPointVertex = packed record
    v: TD3DXVector3;
    color: TD3DColor;
  end;
  PPointVertexArray = ^TPointVertexArray;
  TPointVertexArray = array [0..0] of TPointVertex;

const
  TPointVertex_FVF = (D3DFVF_XYZ or D3DFVF_DIFFUSE);

//-----------------------------------------------------------------------------
// Global structs and data for the ground object
//-----------------------------------------------------------------------------
const
  GROUND_WIDTH          = 256.0;
  GROUND_HEIGHT         = 256.0;
  GROUND_GRIDSIZE       = 8;
  GROUND_TILE           = 32;
  GROUND_COLOR          = $cccccccc;

//-----------------------------------------------------------------------------
// Global data for the particles
//-----------------------------------------------------------------------------
type
  PParticle = ^TParticle;
  TParticle = packed record
    m_bSpark:   Boolean;      // Sparks are less energetic particles that
                              // are generated where/when the main particles
                              // hit the ground

    m_vPos:     TD3DXVector3;  // Current position
    m_vVel:     TD3DXVector3;  // Current velocity

    m_vPos0:    TD3DXVector3; // Initial position
    m_vVel0:    TD3DXVector3; // Initial velocity
    m_fTime0:   Single;       // Time of creation

    m_clrDiffuse:       TD3DXColor;  // Initial diffuse color
    m_clrFade:          TD3DXColor;  // Faded diffuse color
    m_fFade:            Single;      // Fade progression

    m_pNext: PParticle;      // Next particle in list
  end;

  TParticleColors = (COLOR_WHITE, COLOR_RED, COLOR_GREEN, COLOR_BLUE{, NUM_COLORS});

var
  g_clrColor: array[COLOR_WHITE..COLOR_BLUE] of TD3DXColor =
  (
    (r: 1.0;   g: 1.0; b: 1.0; a: 1.0),
    (r: 1.0;   g: 0.5; b: 0.5; a: 1.0),
    (r: 0.5;   g: 0.5; b: 0.5; a: 1.0),
    (r: 0.125; g: 0.5; b: 1.0; a: 1.0)
  );

  g_clrColorFade: array[COLOR_WHITE..COLOR_BLUE] of TD3DXColor =
  (
    (r: 1.0;   g: 0.25; b: 0.25; a: 1.0),
    (r: 1.0;   g: 0.25; b: 0.25; a: 1.0),
    (r: 0.25;  g: 0.75; b: 0.25; a: 1.0),
    (r: 0.125; g: 0.25; b: 0.75; a: 1.0)
  );

  //-----------------------------------------------------------------------------
  // Name:
  // Desc:
  //-----------------------------------------------------------------------------
type
  CParticleSystem = class
  protected
    m_fRadius:  Single;

    m_dwBase:          DWORD;
    m_dwFlush:         DWORD;
    m_dwDiscard:       DWORD;

    m_dwParticles:     DWORD;
    m_dwParticlesLim:  DWORD;
    m_pParticles:      PParticle;
    m_pParticlesFree:  PParticle;

    // Geometry
    m_pVB:             IDirect3DVertexBuffer9;

  public
    constructor Create(dwFlush, dwDiscard: DWord;fRadius: Single);
    destructor Destroy; override;

    function RestoreDeviceObjects(pd3dDevice: IDirect3DDevice9): HResult;
    function InvalidateDeviceObjects: HResult;

    function Update(fSecsPerFrame: Single; dwNumParticlesToEmit: DWord;
                    const clrEmitColor, clrFadeColor: TD3DXColor;
                    fEmitVel: Single; vPosition: TD3DXVector3): HResult;

    function Render(pd3dDevice: IDirect3DDevice9): HResult;
  end;

  CMyD3DApplication = class(CD3DApplication)
  private
    m_pFont:            CD3DFont;
    m_pFontSmall:       CD3DFont;

    // Ground stuff
    m_pGroundTexture:   IDirect3DTexture9;

    m_planeGround:      TD3DXPlane;

    m_pGroundVB:            IDirect3DVertexBuffer9;
    m_pGroundIB:            IDirect3DIndexBuffer9;
    m_dwNumGroundIndices:   DWORD;
    m_dwNumGroundVertices:  DWORD;

    // Particle stuff
    m_pParticleTexture:     IDirect3DTexture9;
    m_pParticleSystem:      CParticleSystem;
    m_dwNumParticlesToEmit: DWORD;
    m_dwParticleColor:      TParticleColors;
    m_bAnimateEmitter:      Boolean;

    m_bKey:                 array[0..255] of Byte;
    m_bDrawReflection:      Boolean;
    m_bCanDoAlphaBlend:     Boolean;
    m_bShowHelp:            Boolean;

    // Variables for determining view position
    m_vPosition:            TD3DXVector3;
    m_vVelocity:            TD3DXVector3;
    m_fYaw:                 Single;
    m_fYawVelocity:         Single;
    m_fPitch:               Single;
    m_fPitchVelocity:       Single;
    m_matView:              TD3DXMatrix;
    m_matOrientation:       TD3DXMatrix;

  protected
    function OneTimeSceneInit: HResult; override;
    function InitDeviceObjects: HResult; override;
    function RestoreDeviceObjects: HResult; override;
    function InvalidateDeviceObjects: HResult; override;
    function DeleteDeviceObjects: HResult; override;
    function Render: HResult; overload; override;
    function FrameMove: HResult; override;
    function FinalCleanup: HResult; override;

    procedure UpdateCamera;
  public
    constructor Create;
    function MsgProc(hWnd: HWND; uMsg: UINT; wParam: WPARAM; lParam: LPARAM): LRESULT; override;

    function ConfirmDevice(const pCaps: TD3DCaps9; dwBehavior: DWORD;
      adapterFormat: TD3DFormat; backBufferFormat: TD3DFormat): HRESULT; override;
 end;

implementation

function FtoDW(f: Single): DWORD;
begin
  Result:= PDWord(@f)^;
end;

{ CMyD3DApplication }

//-----------------------------------------------------------------------------
// Name: CMyD3DApplication()
// Desc: Application constructor. Sets attributes for the app.
//-----------------------------------------------------------------------------
constructor CMyD3DApplication.Create;
begin
  inherited;

  m_strWindowTitle    := 'PointSprites: Using particle effects';
  m_d3dEnumeration.AppUsesDepthBuffer := True;

  m_pFont           := CD3DFont.Create('Arial', 12, D3DFONT_BOLD);
  m_pFontSmall      := CD3DFont.Create('Arial',  9, D3DFONT_BOLD);

  m_pParticleSystem      := CParticleSystem.Create(512, 2048, 0.03);
  m_pParticleTexture     := nil;
  m_dwNumParticlesToEmit := 10;
  m_bAnimateEmitter      := False;
  m_dwParticleColor      := COLOR_WHITE;

  m_pGroundTexture       := nil;
  m_dwNumGroundVertices  := (GROUND_GRIDSIZE + 1) * (GROUND_GRIDSIZE + 1);
  m_dwNumGroundIndices   := (GROUND_GRIDSIZE * GROUND_GRIDSIZE) * 6;
  m_pGroundVB            := nil;
  m_pGroundIB            := nil;
  m_planeGround          := D3DXPlane(0.0, 1.0, 0.0, 0.0);

  m_bDrawReflection      := False;
  m_bCanDoAlphaBlend     := False;
  m_bShowHelp            := False;

  ZeroMemory(@m_bKey, 256);
  m_vPosition      := D3DXVector3(0.0, 3.0,-4.0);
  m_vVelocity      := D3DXVector3(0.0, 0.0, 0.0);
  m_fYaw           := 0.03;
  m_fYawVelocity   := 0.0;
  m_fPitch         := 0.5;
  m_fPitchVelocity := 0.0;
  D3DXMatrixTranslation(m_matView, 0.0, 0.0, 10.0);
  D3DXMatrixTranslation(m_matOrientation, 0.0, 0.0, 0.0);
end;

//-----------------------------------------------------------------------------
// Name: OneTimeSceneInit()
// Desc: Called during initial app startup, this function performs all the
//       permanent initialization.
//-----------------------------------------------------------------------------
function CMyD3DApplication.OneTimeSceneInit: HResult;
begin
  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: FrameMove()
// Desc: Called once per frame, the call is the entry point for animating
//       the scene.
//-----------------------------------------------------------------------------
function CMyD3DApplication.FrameMove: HResult;
var
  vEmitterPostion: TD3DXVector3;
begin
  // Slow things down for the REF device
  if (m_d3dCaps.DeviceType = D3DDEVTYPE_REF) then m_fElapsedTime:= 0.05;

  // Determine emitter position
  vEmitterPostion:= D3DXVector3Zero;
  if (m_bAnimateEmitter) then
    vEmitterPostion:= D3DXVector3(3*Sin(m_fTime), 0.0, 3*Cos(m_fTime));

  // Update particle system
  m_pParticleSystem.Update(m_fElapsedTime, m_dwNumParticlesToEmit,
                           g_clrColor[m_dwParticleColor],
                           g_clrColorFade[m_dwParticleColor], 8.0,
                           vEmitterPostion);

  Result := S_OK;
end;

//-----------------------------------------------------------------------------
// Name: Render()
// Desc: Called once per frame, the call is the entry point for 3d
//       rendering. This function sets up render states, clears the
//       viewport, and renders the scene.
//-----------------------------------------------------------------------------
function CMyD3DApplication.Render: HResult;
var
  matReflectedView: TD3DXMatrixA16;
begin
  // Update the camera here rather than in FrameMove so you can
  // move the camera even when the scene is paused
  UpdateCamera;

  if FAILED(m_pd3dDevice.BeginScene) then
  begin
    Result:= S_OK;
    Exit;
  end;

  // Clear the viewport
  m_pd3dDevice.Clear(0, nil, D3DCLEAR_TARGET or D3DCLEAR_ZBUFFER, $00, 1.0, 0);

  // Draw reflection of particles
  if m_bDrawReflection then
  begin
    D3DXMatrixReflect (matReflectedView, m_planeGround);
    D3DXMatrixMultiply(matReflectedView, matReflectedView, m_matView);

    m_pd3dDevice.SetRenderState(D3DRS_ZWRITEENABLE, iFalse);
    m_pd3dDevice.SetRenderState(D3DRS_ALPHABLENDENABLE, iTrue);
    m_pd3dDevice.SetRenderState(D3DRS_SRCBLEND, D3DBLEND_ONE);
    m_pd3dDevice.SetRenderState(D3DRS_DESTBLEND, D3DBLEND_ONE);

    m_pd3dDevice.SetTransform(D3DTS_VIEW, matReflectedView);
    m_pd3dDevice.SetTexture(0, m_pParticleTexture);
    m_pParticleSystem.Render(m_pd3dDevice);

    m_pd3dDevice.SetRenderState(D3DRS_ZWRITEENABLE, iTrue);
    m_pd3dDevice.SetRenderState(D3DRS_ALPHABLENDENABLE, iFalse);
  end;


  // Draw the ground
  if m_bDrawReflection then
  begin
    m_pd3dDevice.SetRenderState(D3DRS_ALPHABLENDENABLE, iTrue);
    m_pd3dDevice.SetRenderState(D3DRS_SRCBLEND, D3DBLEND_SRCALPHA);
    m_pd3dDevice.SetRenderState(D3DRS_DESTBLEND, D3DBLEND_INVSRCALPHA);
  end;

  m_pd3dDevice.SetTransform(D3DTS_VIEW, m_matView);
  m_pd3dDevice.SetTexture(0, m_pGroundTexture);
  m_pd3dDevice.SetFVF(TColorVertex_FVF);
  m_pd3dDevice.SetStreamSource(0, m_pGroundVB, 0, SizeOf(TColorVertex));
  m_pd3dDevice.SetIndices(m_pGroundIB);
  m_pd3dDevice.DrawIndexedPrimitive(D3DPT_TRIANGLELIST, 0, 0, m_dwNumGroundVertices,
    0, m_dwNumGroundIndices div 3);


  // Draw particles
  m_pd3dDevice.SetRenderState(D3DRS_ZWRITEENABLE, iFalse);
  m_pd3dDevice.SetRenderState(D3DRS_ALPHABLENDENABLE, iTrue);
  m_pd3dDevice.SetRenderState(D3DRS_SRCBLEND,  D3DBLEND_ONE);
  m_pd3dDevice.SetRenderState(D3DRS_DESTBLEND, D3DBLEND_ONE);

  m_pd3dDevice.SetTexture(0, m_pParticleTexture);
  m_pParticleSystem.Render(m_pd3dDevice);

  m_pd3dDevice.SetRenderState(D3DRS_ZWRITEENABLE, iTrue);
  m_pd3dDevice.SetRenderState(D3DRS_ALPHABLENDENABLE, iFalse);


  // Output statistics
  m_pFont.DrawText( 2,  0, D3DCOLOR_ARGB(255,255,255,0), m_strFrameStats);
  m_pFont.DrawText( 2, 20, D3DCOLOR_ARGB(255,255,255,0), m_strDeviceStats);


  // Draw help
  if m_bShowHelp then
  begin
    m_pFontSmall.DrawText(2, 40, $ffffffff,
                          'Keyboard controls:');
    m_pFontSmall.DrawText(20, 60, $ffffffff,
                          'Move'#10'Turn'#10'Pitch'#10'Slide'#10#10'Help'#10 +
                          'Change device'#10'Animate emitter'#10 +
                          'Change color'#10#10'Toggle reflection'#10 +
                          'Exit');
    m_pFontSmall.DrawText(160, 60, $ffffffff,
                          'W,S'#10'E,Q'#10'A,Z'#10'Arrow keys'#10#10 +
                          'F1'#10'F2'#10'F3'#10'F4'#10#10'R'#10'Esc');
  end
  else
    m_pFontSmall.DrawText(2, 40, $ffffffff,
                          'Press F1 for help');


  // End the scene.
  m_pd3dDevice.EndScene;
  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: UpdateCamera()
// Desc:
//-----------------------------------------------------------------------------
procedure CMyD3DApplication.UpdateCamera;
var
  fElapsedTime: Single;
  fSpeed: Single;
  fAngularSpeed: Single;

  vT: TD3DXVector3;
  qR: TD3DXQuaternion;
begin
  if (m_fElapsedTime > 0.0)
    then fElapsedTime:= m_fElapsedTime
    else fElapsedTime:= 0.05;

  fSpeed        := 3.0*fElapsedTime;
  fAngularSpeed := 1.0*fElapsedTime;

  // De-accelerate the camera movement (for smooth motion)
  D3DXVec3Scale(m_vVelocity, m_vVelocity, 0.9);
  m_fYawVelocity:= m_fYawVelocity   * 0.9;
  m_fPitchVelocity:= m_fPitchVelocity * 0.9;

  // Process keyboard input
  if (0 <> m_bKey[VK_RIGHT])    then m_vVelocity.x    := m_vVelocity.x + fSpeed; // Slide Right
  if (0 <> m_bKey[VK_LEFT])     then m_vVelocity.x    := m_vVelocity.x - fSpeed; // Slide Left
  if (0 <> m_bKey[VK_UP])       then m_vVelocity.y    := m_vVelocity.y + fSpeed; // Slide Up
  if (0 <> m_bKey[VK_DOWN])     then m_vVelocity.y    := m_vVelocity.y - fSpeed; // Slide Down
  if (0 <> m_bKey[Ord('W')])    then m_vVelocity.z    := m_vVelocity.z + fSpeed; // Move Forward
  if (0 <> m_bKey[Ord('S')])    then m_vVelocity.z    := m_vVelocity.z - fSpeed; // Move Backward
  if (0 <> m_bKey[Ord('E')])    then m_fYawVelocity   := m_fYawVelocity + fSpeed; // Turn Right
  if (0 <> m_bKey[Ord('Q')])    then m_fYawVelocity   := m_fYawVelocity - fSpeed; // Turn Left
  if (0 <> m_bKey[Ord('Z')])    then m_fPitchVelocity := m_fPitchVelocity + fSpeed; // Turn Down
  if (0 <> m_bKey[Ord('A')])    then m_fPitchVelocity := m_fPitchVelocity - fSpeed; // Turn Up
  if (0 <> m_bKey[VK_ADD])      then if (m_dwNumParticlesToEmit < 10) then Inc(m_dwNumParticlesToEmit);
  if (0 <> m_bKey[VK_SUBTRACT]) then if (m_dwNumParticlesToEmit > 0)  then Dec(m_dwNumParticlesToEmit);

  // Update the position vector
  D3DXVec3Scale(vT, m_vVelocity, fSpeed);
  D3DXVec3TransformNormal(vT, vT, m_matOrientation);
  D3DXVec3Add(m_vPosition, m_vPosition, vT);
  if (m_vPosition.y < 1.0) then m_vPosition.y := 1.0;

  // Update the yaw-pitch-rotation vector
  m_fYaw   := m_fYaw + fAngularSpeed * m_fYawVelocity;
  m_fPitch := m_fPitch + fAngularSpeed * m_fPitchVelocity;
  if (m_fPitch < 0.0)       then m_fPitch:= 0.0;
  if (m_fPitch > D3DX_PI/2) then m_fPitch:= D3DX_PI/2;

  // Set the view matrix
  D3DXQuaternionRotationYawPitchRoll(qR, m_fYaw, m_fPitch, 0.0);
  D3DXMatrixAffineTransformation(m_matOrientation, 1.25, nil, @qR, @m_vPosition);
  D3DXMatrixInverse(m_matView, nil, m_matOrientation);
end;

//-----------------------------------------------------------------------------
// Name: InitDeviceObjects()
// Desc: Initialize scene objects.
//-----------------------------------------------------------------------------
function CMyD3DApplication.InitDeviceObjects: HResult;
var
  strPath: array [0..MAX_PATH-1] of Char;
  pVertices: PColorVertex;
  zz, xx: DWord;
  pIndices: PWord;
  z, x: DWord;
  vtx: DWord;
begin
  // Create textures
  Result := D3DAPPERR_MEDIANOTFOUND;

  // Get the path to the texture
  if FAILED(DXUtil_FindMediaFileCb(strPath, SizeOf(strPath), 'Ground2.bmp')) then Exit;

  // Create textures
  if FAILED(D3DXCreateTextureFromFileEx(m_pd3dDevice, strPath,
              D3DX_DEFAULT, D3DX_DEFAULT, D3DX_DEFAULT, 0, D3DFMT_UNKNOWN,
              D3DPOOL_MANAGED, D3DX_FILTER_TRIANGLE or D3DX_FILTER_MIRROR,
              D3DX_FILTER_TRIANGLE or D3DX_FILTER_MIRROR, 0, nil, nil, m_pGroundTexture))
  then Exit;

  // Get the path to the texture
  if FAILED(DXUtil_FindMediaFileCb(strPath, SizeOf(strPath), 'Particle.bmp')) then Exit;
  // Create textures
  if FAILED(D3DXCreateTextureFromFileEx( m_pd3dDevice, strPath,
              D3DX_DEFAULT, D3DX_DEFAULT, D3DX_DEFAULT, 0, D3DFMT_UNKNOWN,
              D3DPOOL_MANAGED, D3DX_FILTER_TRIANGLE or D3DX_FILTER_MIRROR,
              D3DX_FILTER_TRIANGLE or D3DX_FILTER_MIRROR, 0, nil, nil, m_pParticleTexture))
  then Exit;

  // Set up the fonts and textures
  m_pFont.InitDeviceObjects(m_pd3dDevice);
  m_pFontSmall.InitDeviceObjects(m_pd3dDevice);

  // Check if we can do the reflection effect
  m_bCanDoAlphaBlend:= ((m_d3dCaps.SrcBlendCaps  and D3DPBLENDCAPS_SRCALPHA) <> 0) and
                       ((m_d3dCaps.DestBlendCaps and D3DPBLENDCAPS_INVSRCALPHA) <> 0);

  if (m_bCanDoAlphaBlend) then m_bDrawReflection := TRUE;

  // Create ground object
  begin
    // Create vertex buffer for ground object
    Result:= m_pd3dDevice.CreateVertexBuffer(m_dwNumGroundVertices*SizeOf(TColorVertex),
                            D3DUSAGE_WRITEONLY, TColorVertex_FVF, D3DPOOL_MANAGED, m_pGroundVB, nil);
    if FAILED(Result) then Exit;

    // Fill vertex buffer
    Result:= m_pGroundVB.Lock(0, 0, Pointer(pVertices), 0);
    if FAILED(Result) then Exit;

    // Fill in vertices
    for zz:= 0 to GROUND_GRIDSIZE do
    begin
      for xx := 0 to GROUND_GRIDSIZE do
      begin
        pVertices.v.x   := GROUND_WIDTH * (xx/GROUND_GRIDSIZE-0.5);
        pVertices.v.y   := 0.0;
        pVertices.v.z   := GROUND_HEIGHT * (zz/GROUND_GRIDSIZE-0.5);
        pVertices.color := GROUND_COLOR;
        pVertices.tu    := xx*GROUND_TILE/GROUND_GRIDSIZE;
        pVertices.tv    := zz*GROUND_TILE/GROUND_GRIDSIZE;
        Inc(pVertices);
      end;
    end;

    m_pGroundVB.Unlock;

    // Create the index buffer
    Result:= m_pd3dDevice.CreateIndexBuffer(m_dwNumGroundIndices*SizeOf(Word),
                            D3DUSAGE_WRITEONLY, D3DFMT_INDEX16, D3DPOOL_MANAGED,
                            m_pGroundIB, nil);
    if FAILED(Result) then Exit;

    // Fill the index buffer
    Result:= m_pGroundIB.Lock(0, m_dwNumGroundIndices*SizeOf(Word), Pointer(pIndices), 0);
    if FAILED(Result) then
    begin
      Result:= E_FAIL;
      Exit;
    end;

    // Fill in indices
    for z:= 0 to GROUND_GRIDSIZE - 1 do
    begin
      for x:= 0 to GROUND_GRIDSIZE - 1 do
      begin
        vtx:= x + z * (GROUND_GRIDSIZE+1);
        pIndices^:= (vtx + 1);                          Inc(pIndices);
        pIndices^:= (vtx + 0);                          Inc(pIndices);
        pIndices^:= (vtx + 0 + (GROUND_GRIDSIZE+1));    Inc(pIndices);
        pIndices^:= (vtx + 1);                          Inc(pIndices);
        pIndices^:= (vtx + 0 + (GROUND_GRIDSIZE+1));    Inc(pIndices);
        pIndices^:= (vtx + 1 + (GROUND_GRIDSIZE+1));    Inc(pIndices);
      end;
    end;

    m_pGroundIB.Unlock;
  end;

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: RestoreDeviceObjects()
// Desc:
//-----------------------------------------------------------------------------
function CMyD3DApplication.RestoreDeviceObjects: HResult;
var
  matWorld, matProj: TD3DXMatrixA16;
  fAspect: Single;
begin
  m_pFont.RestoreDeviceObjects;
  m_pFontSmall.RestoreDeviceObjects;

  // Set the world matrix
  D3DXMatrixIdentity(matWorld);
  m_pd3dDevice.SetTransform(D3DTS_WORLD, matWorld);

  // Set projection matrix
  fAspect := (m_d3dsdBackBuffer.Width) / m_d3dsdBackBuffer.Height;
  D3DXMatrixPerspectiveFovLH(matProj, D3DX_PI/4, fAspect, 0.1, 100.0);
  m_pd3dDevice.SetTransform(D3DTS_PROJECTION, matProj);

  m_pd3dDevice.SetSamplerState(0, D3DSAMP_MINFILTER, D3DTEXF_LINEAR);
  m_pd3dDevice.SetSamplerState(0, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR);
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_COLOROP,   D3DTOP_MODULATE);
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_ALPHAOP,   D3DTOP_MODULATE);
  m_pd3dDevice.SetTextureStageState(1, D3DTSS_COLOROP,   D3DTOP_DISABLE);
  m_pd3dDevice.SetTextureStageState(1, D3DTSS_ALPHAOP,   D3DTOP_DISABLE);

  m_pd3dDevice.SetRenderState(D3DRS_SRCBLEND,  D3DBLEND_ONE);
  m_pd3dDevice.SetRenderState(D3DRS_DESTBLEND, D3DBLEND_ONE);
  m_pd3dDevice.SetRenderState(D3DRS_LIGHTING,  iFalse);
  m_pd3dDevice.SetRenderState(D3DRS_CULLMODE,  D3DCULL_CCW);
  m_pd3dDevice.SetRenderState(D3DRS_SHADEMODE, D3DSHADE_FLAT);

  // Initialize the particle system
  Result:= m_pParticleSystem.RestoreDeviceObjects(m_pd3dDevice);
  if FAILED(Result) then Exit;

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: InvalidateDeviceObjects()
// Desc:
//-----------------------------------------------------------------------------
function CMyD3DApplication.InvalidateDeviceObjects: HResult;
begin
  m_pFont.InvalidateDeviceObjects;
  m_pFontSmall.InvalidateDeviceObjects;
  m_pParticleSystem.InvalidateDeviceObjects;

  Result := S_OK;
end;

//-----------------------------------------------------------------------------
// Name: DeleteDeviceObjects()
// Desc: Called when the app is exiting, or the device is being changed,
//       this function deletes any device dependent objects.
//-----------------------------------------------------------------------------
function CMyD3DApplication.DeleteDeviceObjects: HResult;
begin
  m_pFont.DeleteDeviceObjects;
  m_pFontSmall.DeleteDeviceObjects;

  m_pGroundTexture:= nil;
  m_pParticleTexture:= nil;

  m_pGroundVB:= nil;
  m_pGroundIB:= nil;

  Result := S_OK;
end;

//-----------------------------------------------------------------------------
// Name: FinalCleanup()
// Desc: Called before the app exits, this function gives the app the chance
//       to cleanup after itself.
//-----------------------------------------------------------------------------
function CMyD3DApplication.FinalCleanup: HResult;
begin
  FreeAndNil(m_pFont);
  FreeAndNil(m_pFontSmall);
  FreeAndNil(m_pGroundTexture);
  FreeAndNil(m_pParticleTexture);
  FreeAndNil(m_pParticleSystem);

  Result := S_OK;
end;

//-----------------------------------------------------------------------------
// Name: ConfirmDevice()
// Desc: Called during device initialization, this code checks the device
//       for some minimum set of capabilities
//-----------------------------------------------------------------------------
function CMyD3DApplication.ConfirmDevice(const pCaps: TD3DCaps9; dwBehavior: DWORD;
  adapterFormat: TD3DFormat; backBufferFormat: TD3DFormat): HRESULT; 
begin
  Result:= E_FAIL;

  // Make sure device can do ONE:ONE alphablending
  if (0 = (pCaps.SrcBlendCaps  and D3DPBLENDCAPS_ONE)) then Exit;
  if (0 = (pCaps.DestBlendCaps and D3DPBLENDCAPS_ONE)) then Exit;

  // Make sure HW TnL devices can do point sprites
  if ((dwBehavior and D3DCREATE_HARDWARE_VERTEXPROCESSING) or
      (dwBehavior and D3DCREATE_MIXED_VERTEXPROCESSING)) <> 0 then
  begin
    if (pCaps.MaxPointSize <= 1.0) then Exit;
  end;

  // Need to support post-pixel processing (for alpha blending)
  if FAILED(m_pD3D.CheckDeviceFormat(pCaps.AdapterOrdinal, pCaps.DeviceType,
       adapterFormat, D3DUSAGE_RENDERTARGET or D3DUSAGE_QUERY_POSTPIXELSHADER_BLENDING,
       D3DRTYPE_SURFACE, backBufferFormat))
  then Exit;

  Result := S_OK;
end;

//-----------------------------------------------------------------------------
// Name: MsgProc()
// Desc: Message proc function to handle key and menu input
//-----------------------------------------------------------------------------
function CMyD3DApplication.MsgProc(hWnd: HWND; uMsg: UINT; wParam: WPARAM;
  lParam: LPARAM): LRESULT;
begin
  // Record key presses
  if (WM_KEYDOWN = uMsg) then
  begin
    m_bKey[wParam]:= 1;
  end;

  // Perform commands when keys are rleased
  if (WM_KEYUP = uMsg) then
  begin
    m_bKey[wParam] := 0;

    case wParam of
      VK_F1:
      begin
        m_bShowHelp := not m_bShowHelp;
        Result := 1;
        Exit;
      end;

      Ord('R'): if m_bCanDoAlphaBlend then m_bDrawReflection := not m_bDrawReflection;
      VK_F3:    m_bAnimateEmitter := not m_bAnimateEmitter;

      VK_F4:
      begin
        if (m_dwParticleColor = High(TParticleColors)) then m_dwParticleColor:= COLOR_WHITE
        else Inc(m_dwParticleColor);
      end;
    end;
  end;

  Result:= inherited MsgProc(hWnd, uMsg, wParam, lParam);
end;

{ CParticleSystem }

//-----------------------------------------------------------------------------
// Name:
// Desc:
//-----------------------------------------------------------------------------
constructor CParticleSystem.Create(dwFlush, dwDiscard: DWord; fRadius: Single);
begin
  m_fRadius        := fRadius;

  m_dwBase         := dwDiscard;
  m_dwFlush        := dwFlush;
  m_dwDiscard      := dwDiscard;

  m_dwParticles    := 0;
  m_dwParticlesLim := 2048;

  m_pParticles     := nil;
  m_pParticlesFree := nil;
  m_pVB            := nil;
end;

//-----------------------------------------------------------------------------
// Name:
// Desc:
//-----------------------------------------------------------------------------
destructor CParticleSystem.Destroy;
var
  pSpark: PParticle;
begin
  InvalidateDeviceObjects;

  while Assigned(m_pParticles) do
  begin
    pSpark := m_pParticles;
    m_pParticles := pSpark.m_pNext;
    FreeMem(pSpark);
  end;

  while Assigned(m_pParticlesFree) do
  begin
    pSpark := m_pParticlesFree;
    m_pParticlesFree := pSpark.m_pNext;
    Dispose(pSpark);
  end;

  inherited;
end;

//-----------------------------------------------------------------------------
// Name:
// Desc:
//-----------------------------------------------------------------------------
function CParticleSystem.RestoreDeviceObjects(pd3dDevice: IDirect3DDevice9): HResult;
begin
  // Create a vertex buffer for the particle system.  The size of this buffer
  // does not relate to the number of particles that exist.  Rather, the
  // buffer is used as a communication channel with the device.. we fill in
  // a bit, and tell the device to draw.  While the device is drawing, we
  // fill in the next bit using NOOVERWRITE.  We continue doing this until
  // we run out of vertex buffer space, and are forced to DISCARD the buffer
  // and start over at the beginning.

  Result:= pd3dDevice.CreateVertexBuffer(m_dwDiscard * SizeOf(TPointVertex),
              D3DUSAGE_DYNAMIC or D3DUSAGE_WRITEONLY or D3DUSAGE_POINTS,
              TPointVertex_FVF, D3DPOOL_DEFAULT, m_pVB, nil);
  if FAILED(Result) then
  begin
    Result := E_FAIL;
    Exit;
  end;

  Result := S_OK;
end;

//-----------------------------------------------------------------------------
// Name:
// Desc:
//-----------------------------------------------------------------------------
function CParticleSystem.InvalidateDeviceObjects: HResult;
begin
  m_pVB:= nil;
  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name:
// Desc:
//-----------------------------------------------------------------------------
function CParticleSystem.Update(fSecsPerFrame: Single;
  dwNumParticlesToEmit: DWord; const clrEmitColor, clrFadeColor: TD3DXColor;
  fEmitVel: Single; vPosition: TD3DXVector3): HResult;
{$WRITEABLECONST ON}
const
  fTime: Single = 0.0;
{$WRITEABLECONST OFF}
var
  p_Particle: PParticle;
  ppParticle: ^PParticle;

  fT: Single;
  fGravity: Single;

  v1: TD3DXVector3;

  i: Integer;
  pSpark: PParticle;

  fRand1: Single;
  fRand2: Single;

  dwParticlesEmit: DWord;
begin
  pSpark:= nil;
  p_Particle:= nil;
  fTime:= fTime + fSecsPerFrame;

  ppParticle:= @m_pParticles;

  while Assigned(ppParticle^) do
  begin
    p_Particle := ppParticle^;

    // Calculate new position
    fT := fTime - p_Particle.m_fTime0;

    if (p_Particle.m_bSpark) then
    begin
      fGravity := -5.0;
      p_Particle.m_fFade:= p_Particle.m_fFade - fSecsPerFrame * 2.25;
    end else
    begin
      fGravity := -9.8;
      p_Particle.m_fFade:= p_Particle^.m_fFade - fSecsPerFrame * 0.25;
    end;

    // p_Particle.m_vPos    := p_Particle.m_vVel0 * fT + p_Particle.m_vPos0;
    D3DXVec3Add(p_Particle.m_vPos, D3DXVec3Scale(v1, p_Particle.m_vVel0, fT)^, p_Particle.m_vPos0);
    p_Particle.m_vPos.y  := p_Particle.m_vPos.y + (0.5 * fGravity) * (fT * fT);
    p_Particle.m_vVel.y  := p_Particle.m_vVel0.y + fGravity * fT;

    if (p_Particle.m_fFade < 0.0) then p_Particle.m_fFade := 0.0;

    // Kill old particles
    if (p_Particle.m_vPos.y < m_fRadius) or
       p_Particle.m_bSpark and (p_Particle.m_fFade <= 0.0) then
    begin
      // Emit sparks
      if not p_Particle.m_bSpark then
      begin
        for i:= 0 to 3 do
        begin
          if Assigned(m_pParticlesFree) then
          begin
            pSpark:= m_pParticlesFree;
            m_pParticlesFree:= pSpark.m_pNext;
          end else
          begin
            try
              New(pSpark);
            except
              Result := E_OUTOFMEMORY;
              Exit;
            end;
          end;

          pSpark.m_pNext := p_Particle.m_pNext;
          p_Particle.m_pNext := pSpark;

          pSpark.m_bSpark  := TRUE;
          pSpark.m_vPos0   := p_Particle.m_vPos;
          pSpark.m_vPos0.y := m_fRadius;

          fRand1 := Random * D3DX_PI * 2.00;
          fRand2 := Random * D3DX_PI * 0.25;

          pSpark.m_vVel0.x  := p_Particle.m_vVel.x * 0.25 + Cos(fRand1) * Sin(fRand2);
          pSpark.m_vVel0.z  := p_Particle.m_vVel.z * 0.25 + Sin(fRand1) * Sin(fRand2);
          pSpark.m_vVel0.y  := Cos(fRand2);
          pSpark.m_vVel0.y  := pSpark.m_vVel0.y * Random * 1.5;

          pSpark.m_vPos := pSpark.m_vPos0;
          pSpark.m_vVel := pSpark.m_vVel0;

          D3DXColorLerp(pSpark.m_clrDiffuse, p_Particle.m_clrFade,
                        p_Particle.m_clrDiffuse, p_Particle.m_fFade);
          pSpark.m_clrFade := D3DXColor(0.0, 0.0, 0.0, 1.0);
          pSpark.m_fFade   := 1.0;
          pSpark.m_fTime0  := fTime;
        end;
      end;

      // Kill particle
      ppParticle^ := p_Particle.m_pNext;
      p_Particle.m_pNext := m_pParticlesFree;
      m_pParticlesFree := p_Particle;

      if (not p_Particle.m_bSpark) then Dec(m_dwParticles);
    end else
    begin
      ppParticle := @p_Particle.m_pNext;
    end;
  end;

  // Emit new particles
  dwParticlesEmit:= m_dwParticles + dwNumParticlesToEmit;
  while (m_dwParticles < m_dwParticlesLim) and (m_dwParticles < dwParticlesEmit) do
  begin
    if Assigned(m_pParticlesFree) then
    begin
      p_Particle := m_pParticlesFree;
      m_pParticlesFree := p_Particle.m_pNext;
    end else
    begin
      try
        New(p_Particle);
      except
        Result := E_OUTOFMEMORY;
        Exit;
      end;
    end;

    p_Particle.m_pNext := m_pParticles;
    m_pParticles := p_Particle;
    Inc(m_dwParticles);

    // Emit new particle
    fRand1 := Random * D3DX_PI * 2.0;
    fRand2 := Random * D3DX_PI * 0.25;

    p_Particle.m_bSpark := FALSE;

    // pParticle.m_vPos0 := vPosition + D3DXVECTOR3(0.0, m_fRadius, 0.0);
    D3DXVec3Add(p_Particle.m_vPos0, vPosition, D3DXVector3(0.0, m_fRadius, 0.0));

    p_Particle.m_vVel0.x  := Cos(fRand1) * Sin(fRand2) * 2.5;
    p_Particle.m_vVel0.z  := Sin(fRand1) * Sin(fRand2) * 2.5;
    p_Particle.m_vVel0.y  := Cos(fRand2);
    p_Particle.m_vVel0.y  := p_Particle.m_vVel0.y * Random * fEmitVel;

    p_Particle.m_vPos := p_Particle.m_vPos0;
    p_Particle.m_vVel := p_Particle.m_vVel0;

    p_Particle.m_clrDiffuse := clrEmitColor;
    p_Particle.m_clrFade    := clrFadeColor;
    p_Particle.m_fFade      := 1.0;
    p_Particle.m_fTime0     := fTime;
  end;

  Result := S_OK;
end;

//-----------------------------------------------------------------------------
// Name: Render()
// Desc: Renders the particle system using either pointsprites (if supported)
//       or using 4 vertices per particle
//-----------------------------------------------------------------------------
function CParticleSystem.Render(pd3dDevice: IDirect3DDevice9): HResult;
var
  p_Particle: PParticle;
  pVertices: PPointVertex;
  dwNumParticlesToRender: DWord;
  Flags: DWord;

  vPos, vVel: TD3DXVector3;
  fLengthSq: Single;
  dwSteps: DWord;

  clrDiffuse: TD3DXColor;
  dwDiffuse: DWORD;
  i: Integer;
begin
  // Set the render states for using point sprites
  pd3dDevice.SetRenderState(D3DRS_POINTSPRITEENABLE, iTrue);
  pd3dDevice.SetRenderState(D3DRS_POINTSCALEENABLE,  iTrue);
  pd3dDevice.SetRenderState(D3DRS_POINTSIZE,     FtoDW(0.08));
  pd3dDevice.SetRenderState(D3DRS_POINTSIZE_MIN, FtoDW(0.00));
  pd3dDevice.SetRenderState(D3DRS_POINTSCALE_A,  FtoDW(0.00));
  pd3dDevice.SetRenderState(D3DRS_POINTSCALE_B,  FtoDW(0.00));
  pd3dDevice.SetRenderState(D3DRS_POINTSCALE_C,  FtoDW(1.00));

  // Set up the vertex buffer to be rendered
  pd3dDevice.SetStreamSource(0, m_pVB, 0, SizeOf(TPointVertex));
  pd3dDevice.SetFVF(TPointVertex_FVF);

  p_Particle := m_pParticles;
  dwNumParticlesToRender := 0;

  // Lock the vertex buffer.  We fill the vertex buffer in small
  // chunks, using D3DLOCK_NOOVERWRITE.  When we are done filling
  // each chunk, we call DrawPrim, and lock the next chunk.  When
  // we run out of space in the vertex buffer, we start over at
  // the beginning, using D3DLOCK_DISCARD.

  Inc(m_dwBase, m_dwFlush);

  if (m_dwBase >= m_dwDiscard) then m_dwBase:= 0;

  if m_dwBase = 0 then Flags:= D3DLOCK_DISCARD else Flags:= D3DLOCK_NOOVERWRITE;

  Result:= m_pVB.Lock(m_dwBase * SizeOf(TPointVertex), m_dwFlush * SizeOf(TPointVertex),
             Pointer(pVertices), Flags);
  if FAILED(Result) then Exit;

  // Render each particle
  while Assigned(p_Particle) do
  begin
    vPos:= p_Particle.m_vPos;
    vVel:= p_Particle.m_vVel;
    fLengthSq := D3DXVec3LengthSq(vVel);

    if (fLengthSq < 1.0)        then dwSteps := 2
    else if (fLengthSq <  4.00) then dwSteps := 3
    else if (fLengthSq <  9.00) then dwSteps := 4
    else if (fLengthSq < 12.25) then dwSteps := 5
    else if (fLengthSq < 16.00) then dwSteps := 6
    else if (fLengthSq < 20.25) then dwSteps := 7
    else                             dwSteps := 8;

    D3DXVec3Scale(vVel, vVel, -0.04 / dwSteps);

    D3DXColorLerp(clrDiffuse, p_Particle.m_clrFade, p_Particle.m_clrDiffuse, p_Particle.m_fFade);
    dwDiffuse:= D3DXColorToDWord(clrDiffuse);

    // Render each particle a bunch of times to get a blurring effect
    for i:= 0 to dwSteps - 1 do
    begin
      pVertices.v     := vPos;
      pVertices.color := dwDiffuse;
      Inc(pVertices);

      Inc(dwNumParticlesToRender);
      if (dwNumParticlesToRender = m_dwFlush) then
      begin
        // Done filling this chunk of the vertex buffer.  Lets unlock and
        // draw this portion so we can begin filling the next chunk.

        m_pVB.Unlock;

        Result:= pd3dDevice.DrawPrimitive(D3DPT_POINTLIST, m_dwBase, dwNumParticlesToRender);
        if FAILED(Result) then Exit;

        // Lock the next chunk of the vertex buffer.  If we are at the
        // end of the vertex buffer, DISCARD the vertex buffer and start
        // at the beginning.  Otherwise, specify NOOVERWRITE, so we can
        // continue filling the VB while the previous chunk is drawing.
        Inc(m_dwBase, m_dwFlush);

        if (m_dwBase >= m_dwDiscard) then m_dwBase:= 0;

        if m_dwBase = 0 then Flags:= D3DLOCK_DISCARD else Flags:= D3DLOCK_NOOVERWRITE;

        Result:= m_pVB.Lock(m_dwBase * SizeOf(TPointVertex), m_dwFlush * SizeOf(TPointVertex),
                    Pointer(pVertices), Flags);
        if FAILED(Result) then Exit;

        dwNumParticlesToRender := 0;
      end;

      D3DXVec3Add(vPos, vPos, vVel);
    end;

    p_Particle := p_Particle.m_pNext;
  end;

  // Unlock the vertex buffer
  m_pVB.Unlock;

  // Render any remaining particles
  if (dwNumParticlesToRender <> 0) then
  begin
    Result:= pd3dDevice.DrawPrimitive(D3DPT_POINTLIST, m_dwBase, dwNumParticlesToRender);
    if FAILED(Result) then Exit;
  end;

  // Reset render states
  pd3dDevice.SetRenderState(D3DRS_POINTSPRITEENABLE, iFalse);
  pd3dDevice.SetRenderState(D3DRS_POINTSCALEENABLE,  iFalse);

  Result := S_OK;
end;

end.

